﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using UnityEditor;
using UnityEditor.Callbacks;
using UnityEngine;

namespace UPP.Engine
{

    public static class Compiler
    {
        public static string LastTimeCompiled { get { return DB.LastCompilation; } }

        private static OGUI.TextBox _logBox = null;
        public static OGUI.TextBox LogBox { get { return _logBox; } set { SetLogBox(value); } }

        private static CompilerDB _db = null;
        public static CompilerDB DB
        {
            get
            {
                if (_db == null)
                    LoadDatabase();
                return _db;
            }
        }

        private static void CreateDatabase()
        {
            _db = ScriptableObject.CreateInstance<CompilerDB>();
            AssetDatabase.CreateAsset(_db, Path.Get(Path.KEY_COMPILER_DB));

            AssetDatabase.Refresh();
            EditorUtility.SetDirty(_db);
            AssetDatabase.SaveAssets();

            _db.LogMemory = new List<string>();
        }

        private static void LoadDatabase()
        {
            _db = AssetDatabase.LoadAssetAtPath<CompilerDB>(Path.Get(Path.KEY_COMPILER_DB));
            if (_db == null)
                CreateDatabase();
        }

        private static void SetLogBox(OGUI.TextBox box)
        {
            _logBox = box;
            if (_logBox != null)
            {
                DB.LogMemory = new List<string>();
                DB.LogMemory.Add(box.Value);
                _logBox.ScrollToEnd();
            }
        }

        private static void Log(string msg)
        {
            DB.LogMemory.Add(msg);
            if (_logBox != null)
            {
                if (_logBox.Value != "")
                    _logBox.Value += n;
                _logBox.Value += msg;
                _logBox.ScrollToEnd();
            }

            if (LogBox != null && LogBox.WindowParent != null)
                (LogBox.WindowParent as EditorWindow).Repaint();
        }

        private static void LogClear()
        {
            DB.LogMemory = new List<string>();
            if (_logBox != null)
                _logBox.Value = "";
        }

        private static string HeaderFill(string headerTemplate, ClassRecord cls)
        {
            string str = headerTemplate;

            str = str.Replace("[class]", cls.className);
            str = str.Replace("[machine]", cls.machineName);
            str = str.Replace("[description]", cls.description);
            str = str.Replace("[abstract]", cls.isAbstract ? "Yes" : "No");
            str = str.Replace("[machine_parent]", cls.machineParent);

            return str;
        }

        private static string HeaderFill(string headerTemplate, TemplateRecord tmp, ClassRecord cls)
        {
            string str = headerTemplate;

            str = str.Replace("[class]", cls.className);
            str = str.Replace("[template]", tmp.TemplateName);
            str = str.Replace("[machine]", tmp.TemplateMachineName);
            str = str.Replace("[description]", cls.description);
            str = str.Replace("[machine_parent]", tmp.MachineParent);

            return str;
        }

        private static string GetHeader(ClassRecord cls)
        {
            string headerTemplate = Settings.GetOrSet("UPP_Gen_Header",
                "Class Name: [class] ([machine])" + n +
                "Description: [description]" + n + n +
                "This file was generated automatically using the Unity++ extension." + n +
                "Do not change this file in any form! All changes will be reverted." + n + n +
                "It is possible to extend the class using PARTIAL class logic (use different file)."
            );

            return HeaderFill(headerTemplate, cls);
        }

        private static string GetHeader(TemplateRecord tmp, ClassRecord cls)
        {
            string headerTemplate = Settings.GetOrSet("UPP_Gen_Header_Template",
                "Template Name: [template] ([machine])" + n +
                "Inherits From: [class] ([machine_parent])" + n +
                "Description: [description]" + n + n +
                "This file was generated automatically using the Unity++ extension." + n +
                "Do not change this file in any form! All changes will be reverted." + n + n +
                "It is possible to extend the class using PARTIAL class logic (use different file)."
            );

            return HeaderFill(headerTemplate, tmp, cls);
        }

        private static string n { get { return Environment.NewLine; } }
        private static string t(int tabIndent)
        {
            string s = "";
            for (int i = 0; i < tabIndent; i++)
                s += "    ";
            return s;
        }

        private static string GetLibraries()
        {
            string defLib = Settings.GetOrSet("UPP_Gen_DefLib",
                "using System.Collections;" + n +
                "using System.Collections.Generic;" + n +
                "using System;" + n + 
                "using UnityEngine;"
            );

            string extraLib = Settings.GetOrSet("Upp_Gen_ExtraLib", "");

            string lib = defLib;
            if (extraLib != "")
                lib = n + extraLib;

            return lib;
        }

        private static string GetClassDeclarationLine(ClassRecord cls)
        {
            string par = "MonoBehaviour";
            if (cls.machineParent != "")
                par = cls.machineParent;
            /*    string abs = "";
                if (cls.isAbstract)
                    abs = "abstract ";*/
            string abs = "abstract ";
            return "public " + abs + "partial class " + cls.machineName + " : " + par;
        }

        private static string GetClassDeclarationLine(TemplateRecord tmp)
        {
            string par = "MonoBehaviour";
            if (tmp.MachineParent != "")
                par = tmp.MachineParent;
            return "public partial class " + tmp.TemplateMachineName + " : " + par;
        }

        private static void WriteLines(this StreamWriter sw, string lines)
        {
            if (lines == "")
                return;
            string[] lns = lines.Split(new string[] { n, "\r", "\n" }, StringSplitOptions.None);
            foreach (var l in lns)
            {
                sw.WriteLine(l);
            }
        }

        private static void WriteLines(this StreamWriter sw, string lines, int indentLevel)
        {
            if (lines == "")
                return;
            string[] lns = lines.Split(new string[] { n, "\r", "\n" }, StringSplitOptions.None);
            foreach (var l in lns)
            {
                sw.WriteLine(t(indentLevel) + l);
            }
        }

        private static void GenerateAttributes(StreamWriter sw, ClassRecord cls, TemplateRecord tmp = null)
        {

            Log("Compiling attributes.");
            var collection = PropertyManager.PropertiesOfClass(cls.machineName);

            List<PropertyBase> bases = new List<PropertyBase>();
            if (tmp == null)
            {
                foreach (var record in collection)
                {
                    PropertyBase b = PropertyBase.FromSerialized(record.Value.PropertyType, record.Value.PropertySerialized);
                    bases.Add(b);
                }
            }
            else
            {
                foreach (var record in tmp.PropertyRecords)
                {
                    PropertyBase b = PropertyBase.FromSerialized(record.PropertyType, record.PropertySerialized);
                    bases.Add(b);
                }
            }
            string lastCategory = "";
            bases.Sort((x, y) => x.Category != y.Category ? x.Category.CompareTo(y.Category) : x.Name.CompareTo(y.Name));

            foreach (var propBase in bases)
            {
                if (propBase.Category != lastCategory)
                {
                    sw.WriteLine(t(2) + "[Space(10)]");
                    sw.WriteLine(t(2) + "[Header(\""+ propBase.Category + "\")]");
                    sw.WriteLine();
                    lastCategory = propBase.Category;
                }

                sw.WriteLine(t(2) + "/// <summary>");
                sw.WriteLine(t(2) + "/// <strong>" + propBase.Name + "</strong>");
                sw.WriteLine(t(2) + "/// <para>" + propBase.Description + "</para>");
                sw.WriteLine(t(2) + "/// </summary>");
                if (tmp != null)
                    sw.WriteLine(t(2) + "[SerializeField]");
                else
                    sw.WriteLine(t(2) + "[NonSerialized]");
                sw.WriteLines(propBase.GenerateSourceAttributes(), 2);
                sw.WriteLine(t(2) + "[Tooltip(\"" + propBase.Name + ": " + propBase.Description + "\")]");
                sw.WriteLines(propBase.GenerateSourceDeclaration(tmp == null), 2);
                sw.WriteLine();
            }
        }

        private static void GeneratePrefabFile(TemplateRecord tmp)
        {
            string pathTemplates = Path.Get(Path.KEY_OUTPUT_PREFABS);
            string extTemplate = Settings.GetOrSet("UPP_Output_Extension_Prefab", ".prefab");

            string fname = tmp.TemplateMachineName + extTemplate;
            string fpath = pathTemplates + "/" + tmp.MachineParent;

            if (!Directory.Exists(pathTemplates))
                Directory.CreateDirectory(pathTemplates);

            if (!Directory.Exists(fpath))
                Directory.CreateDirectory(fpath);

            string fileLocation = fpath + "/" + fname;
            var emptyObj = PrefabUtility.CreateEmptyPrefab(fileLocation);

            GameObject tempObj = new GameObject();
            tempObj.name = "Unity++/" + tmp.MachineParent + "/" + tmp.TemplateMachineName;
            string s = Settings.GetOrSet("UPP_Output_Namespace", "UPP.Game") +
                "." + tmp.TemplateMachineName;
            Type t = Type.GetType(s);
            tempObj.AddComponent(t);

            PrefabUtility.ReplacePrefab(tempObj, emptyObj, ReplacePrefabOptions.ConnectToPrefab);
            GameObject.DestroyImmediate(tempObj);
        }

        private static string GetNamespace()
        {
            string spc = Settings.GetOrSet("UPP_Output_Namespace", "UPP.Game");
            return "namespace " + spc + " {";
        }

        private static void GenerateTemplateFile(StreamWriter sw, TemplateRecord tmp)
        {
            ClassRecord cls = ClassManager.GetClassByMachineName(tmp.MachineParent);

            sw.WriteLines("/*");
            sw.WriteLines(GetHeader(tmp, cls));
            sw.WriteLines("*/");
            sw.WriteLine();
            sw.WriteLines(GetLibraries());
            sw.WriteLine();
            sw.WriteLines(GetNamespace());
            sw.WriteLine();

            sw.WriteLine(t(1) + "/// <summary>");
            sw.WriteLine(t(1) + "/// <strong>" + tmp.TemplateName + "</strong>");
            sw.WriteLine(t(1) + "/// <para>" + cls.description + "</para>");
            sw.WriteLine(t(1) + "/// </summary>");
            sw.WriteLines(GetClassDeclarationLine(tmp) + " {" + n, 1);
            GenerateAttributes(sw, cls, tmp);
            sw.WriteLines("}", 1);
            sw.WriteLines("}");
        }

        private static void GenerateClassFile(StreamWriter sw, ClassRecord cls)
        {
            sw.WriteLines("/*");
            sw.WriteLines(GetHeader(cls));
            sw.WriteLines("*/");
            sw.WriteLine();
            sw.WriteLines(GetLibraries());
            sw.WriteLine();
            sw.WriteLines(GetNamespace());
            sw.WriteLine();

            sw.WriteLine(t(1) + "/// <summary>");
            sw.WriteLine(t(1) + "/// <strong>" + cls.className + "</strong>");
            sw.WriteLine(t(1) + "/// <para>" + cls.description + "</para>");
            sw.WriteLine(t(1) + "/// </summary>");
            sw.WriteLines(GetClassDeclarationLine(cls) + " {" + n, 1);
            GenerateAttributes(sw, cls);
            sw.WriteLines("}", 1);
            sw.WriteLines("}");
        }

        const string PENDING_PREFABS = "UPP_PENDING_PREFAB_RECREATION";

        [DidReloadScripts]
        private static void CreatePrefabs()
        {
            if (EditorPrefs.HasKey(PENDING_PREFABS))
                EditorPrefs.DeleteKey(PENDING_PREFABS);
            else
                return;

            Log("Detected [DidReloadScripts], continuing rutine.");

            var all = TemplateManager.DB.database;
            foreach (var a in all)
            {
                var tmp = TemplateManager.GetTemplateByMachineName(a.machineName);

                Log("Generating prefab file: " + tmp.TemplateMachineName);
                GeneratePrefabFile(tmp);
                Log("Prefab done.");
            }

            Log("Saving Assets.");
            DB.LastCompilation = DateTime.Now.ToLongDateString() + " - " + DateTime.Now.ToLongTimeString();
            EditorUtility.SetDirty(_db);
            AssetDatabase.SaveAssets();

            Log("Refreshing Asset database.");
            AssetDatabase.Refresh();

            Log("Compilation done at " + DB.LastCompilation);

            EditorUtility.DisplayDialog("Compilation completed", "Compilation is complete.", "OK");
        }

        public static void CompileAll()
        {
            LogClear();
            Log("Compilation start at " + DateTime.Now.ToLongTimeString());

            var allClStr = ClassManager.DB.database;
            var allTmStr = TemplateManager.DB.database;

            string pathClasses = Path.Get(Path.KEY_OUTPUT_CLASSES);
            string extClass = Settings.GetOrSet("UPP_Output_Extension_Class", ".cs");

            string pathTemplates = Path.Get(Path.KEY_OUTPUT_TEMPLATES);
            string extTemplate = Settings.GetOrSet("UPP_Output_Extension_Template", ".cs");

            if (!Directory.Exists(pathClasses))
                Directory.CreateDirectory(pathClasses);
            if (!Directory.Exists(pathTemplates))
                Directory.CreateDirectory(pathTemplates);

            Log("Refreshing Asset database.");
            AssetDatabase.Refresh();

            foreach (var a in allClStr)
            {
                Log("Loading class: " + a.machineName);
                var cls = ClassManager.GetClassByMachineName(a.machineName);

                string fname = cls.machineName + extClass;
                string fpath = pathClasses + "/" + fname;

                var file = File.Open(fpath, FileMode.Create, FileAccess.ReadWrite);
                file.Dispose();
                var writer = new StreamWriter(fpath, false);
                writer.NewLine = n;

                Log("Generating class file: " + fpath);
                GenerateClassFile(writer, cls);
                Log("Closing stream: " + fpath);
                writer.Close();
                writer.Dispose();

                Log("Importing Asset into Asset database.");
                //AssetDatabase.ImportAsset(fpath, ImportAssetOptions.ForceSynchronousImport | ImportAssetOptions.ForceUpdate);
                AssetDatabase.ImportAsset(fpath);
                Log("Class done.");
            }

            Log("Saving Assets.");
            AssetDatabase.SaveAssets();

            Log("Refreshing Asset database.");
            AssetDatabase.Refresh();

            List<TemplateRecord> templates = new List<TemplateRecord>();

            foreach (var a in allTmStr)
            {
                Log("Loading template: " + a.machineName);
                var tmp = TemplateManager.GetTemplateByMachineName(a.machineName);
                templates.Add(tmp);

                string fname = tmp.TemplateMachineName + extTemplate;
                string fdir = pathTemplates + "/" + tmp.MachineParent;

                if (!Directory.Exists(fdir))
                    Directory.CreateDirectory(fdir);

                string fpath = fdir + "/" + fname;

                var file = File.Open(fpath, FileMode.Create, FileAccess.ReadWrite);
                file.Dispose();
                var writer = new StreamWriter(fpath, false);
                writer.NewLine = n;

                Log("Generating template file: " + fpath);
                GenerateTemplateFile(writer, tmp);
                Log("Closing stream: " + fpath);
                writer.Close();
                writer.Dispose();

                Log("Importing Asset into Asset database.");
                //AssetDatabase.ImportAsset(fpath, ImportAssetOptions.ForceSynchronousImport | ImportAssetOptions.ForceUpdate);
                AssetDatabase.ImportAsset(fpath);
                Log("Template done.");
            }

            Log("Saving Assets.");
            AssetDatabase.SaveAssets();

            Log("Refreshing Asset database.");
            AssetDatabase.Refresh();

            EditorPrefs.SetString(PENDING_PREFABS, "TRUE");
            Log("Leaving main rutine, waiting for [DidReloadScripts]...");
        }
    }

}