﻿using System.Collections;
using System.Collections.Generic;
using System;
using UnityEngine;

namespace UPP.OGUI
{
    public abstract class Container
    {
        public abstract int GetRenderPriority();

        private List<Control> Controls;
        public List<Control> GetControls()
        {
            return Controls;
        }

        private Dictionary<int, List<Control>> RenderQueue;
        private List<int> RenderLayers;

        private int ControlCounter = -1;
        private int NextControlNumber()
        {
            ControlCounter++;
            return ControlCounter;
        }

        public void RecomputeRenderQueue()
        {
            RenderQueue = new Dictionary<int, List<Control>>();
            RenderLayers = new List<int>();
            
            // We are already guaranteed to have Controls in the right order.
            // We just need to split it based on priorities (e.g. docks...).
            foreach (Control ctrl in Controls)
            {
                int layer = ctrl.GetRenderPriority();
                if (RenderLayers.IndexOf(layer) != -1)
                {
                    RenderQueue[layer].Add(ctrl);
                }
                else
                {
                    RenderLayers.Add(layer);
                    RenderQueue.Add(layer, new List<Control>());
                    RenderQueue[layer].Add(ctrl);
                }
            }
            RenderLayers.Sort();
        }


        public Container()
        {
            WindowObjectRegister.AddObject(this);
            Controls = new List<Control>();
        }

        private int GetNextOrder()
        {
            int next = 0;
            foreach (Control c in Controls)
            {
                if (c.Order >= next)
                    next = c.Order + 1;
            }
            return next;
        }

        public void RemoveControl(Control control)
        {
            BringToFront(control);
            Controls.Remove(control);

            RecomputeRenderQueue();
        }

        public void AddControl(Control control)
        {
            control.Order = GetNextOrder();
            if (control.Name == "")
                control.Name = control.ControlName() + NextControlNumber().ToString().PadLeft(3, '0');
            control.Parent = this;

            Controls.Add(control);

            RecomputeRenderQueue();
        }

        public void SendToBack (Control control)
        {
            foreach(Control c in Controls)
            {
                if (c.Order < control.Order)
                    c.Order++;
            }
            control.Order = 0;

            // Make sure the control is at the min pos:
            Controls.Remove(control);
            Controls.Insert(0, control);

            RecomputeRenderQueue();
        }

        public void BringToFront(Control control)
        {
            int max = 0;
            foreach (Control c in Controls)
            {
                if (c.Order > control.Order)
                    c.Order--;
                if (c.Order > max)
                    max = c.Order;
            }
            control.Order = max + 1;

            // Make sure the control is at the max pos:
            Controls.Remove(control);
            Controls.Add(control);

            RecomputeRenderQueue();
        }

        public void DrawChildren()
        {
            if (this.Controls.Count == 0)
                return;

            Region drawableArea = this.GetDrawableArea();
            foreach (int layer in RenderLayers)
            {
                foreach (Control ctrl in RenderQueue[layer])
                {
                    ctrl.Draw(ref drawableArea);
                }
            }
        }

        public abstract Region GetDrawableArea();
    }
}