﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEditor;
using UnityEngine;

namespace UPP.OGUI
{
    public class Table : Panel
    {
        const float DOUBLE_CLICK_TIME = 0.33f;

        public List<TableColumn> Columns;
        public List<TableRow> Rows;

        private GUIStyle _styleHeaderCols;
        private GUIStyle _styleHeader;
        private GUIStyle _styleRow;
        private GUIStyle _styleRowCells;
        private GUIStyle _styleRowCellsSelected;
        private GUIStyle _styleTitle;

        private float _headerHeight = 20;

        private int _lastClickedRow = -1;
        double _lastClickTime = -99;

        protected Vector2 scrollPos = new Vector2();

        private int _selectedRow = -1;
        public int SelectedRow { get { return _selectedRow; } }
        public void UnselectRows()
        {
            SelectRow(-1);
        }
        public void SelectRow(int rowIndex)
        {
            if (rowIndex >= Rows.Count || rowIndex == _selectedRow)
                return;
            _selectedRow = rowIndex;
            if (OnRowSelected != null)
                OnRowSelected(this, _selectedRow);
        }
        public delegate void OnRowSelectedHandler(object sender, int row);
        public event OnRowSelectedHandler OnRowSelected;

        public delegate void OnDoubleClickHandler(object sender, int row);
        public event OnDoubleClickHandler OnDoubleClick;

        Texture2D lightTexture;
        Texture2D darkTexture;
        Texture2D selectedTexture;
        Texture2D titleTexture;

        public void Clear()
        {
            Rows.Clear();
            _selectedRow = -1;
        }

        private void Init()
        {
            this.Columns = new List<TableColumn>();
            this.Rows = new List<TableRow>();

            _styleHeaderCols = new GUIStyle(UnityEditor.EditorStyles.toolbarButton);
            _styleHeaderCols.fixedHeight = _headerHeight;
            _styleHeader = new GUIStyle(UnityEditor.EditorStyles.toolbar);
            _styleHeader.fixedHeight = _headerHeight;
            _styleHeader.margin = new RectOffset(0, 0, 0, 0);
            _styleHeader.padding = new RectOffset(0, 0, 0, 0);

            _styleRow = new GUIStyle(UnityEditor.EditorStyles.label);
            _styleRowCells = new GUIStyle(UnityEditor.EditorStyles.label);
            _styleRowCells.margin = new RectOffset(0, 0, 0, 0);
            _styleRowCells.padding = new RectOffset(5, 5, 3, 3);
            _styleRowCells.active.textColor = _styleRowCells.normal.textColor;
            _styleRowCellsSelected = new GUIStyle(_styleRowCells);
            _styleRowCellsSelected.normal.textColor = Color.white;
            _styleRowCellsSelected.hover.textColor = _styleRowCellsSelected.normal.textColor;
            _styleRowCellsSelected.focused.textColor = _styleRowCellsSelected.normal.textColor;
            _styleRowCellsSelected.active.textColor = _styleRowCellsSelected.normal.textColor;
            _styleRow.margin = new RectOffset(0, 0, 0, 0);
            _styleRow.padding = new RectOffset(0, 0, 0, 0);
            _styleTitle = new GUIStyle(_styleRowCells);
            _styleTitle.fontStyle = FontStyle.Bold;
            _styleTitle.fontSize = 14;
            _styleTitle.padding = new RectOffset(5, 5, 8, 1);

            lightTexture = new Texture2D(2, 2);
            ColorSchemes.FillTexture(lightTexture, ColorSchemes.BackgroundColorLighter);
            darkTexture = new Texture2D(2, 2);
            ColorSchemes.FillTexture(darkTexture, ColorSchemes.BackgroundColorDarker);
            selectedTexture = new Texture2D(2, 2);
            ColorSchemes.FillTexture(selectedTexture, ColorSchemes.SelectionColor);
            titleTexture = new Texture2D(2, 2);
            ColorSchemes.FillTexture(titleTexture, ColorSchemes.BackgroundColor);

            this.InheritedContent = DrawInherited;

            this.SizeChanged += Table_SizeChanged;
        }

        public void Redraw()
        {
            if (WindowParent != null)
                (WindowParent as EditorWindow).Repaint();
        }

        private void Table_SizeChanged(object sender, float oldWidth, float oldHeight, float newWidth, float newHeight)
        {
            Redraw();
        }

        public Table()
        {
            Init();
        }

        public Table(string[] captions)
        {
            Init();
            AddColumns(captions);
        }

        public Table(List<string> captions)
        {
            Init();
            AddColumns(captions);
        }

        public void AddColumns(string[] captions)
        {
            foreach (string s in captions)
            {
                AddColumn(s);
            }
        }

        public void AddColumns(List<string> captions)
        {
            AddColumns(captions.ToArray());
        }

        public void AddColumn(string caption)
        {
            TableColumn col = new TableColumn(_styleHeaderCols);
            col.Caption = caption;
            Columns.Add(col);
            foreach (TableRow r in this.Rows)
                r.ExpectedColCount = this.Columns.Count;
        }

        public float InsideWidth()
        {
            float w = 0;
            foreach (TableColumn col in Columns)
            {
                w += (col.Width - col.Style.padding.left - col.Style.padding.right);
            }
            return w;
        }

        public float InsideHeight()
        {
            float h = _headerHeight;
            foreach (TableRow row in Rows)
            {
                h += (row.Height);
            }
            return h;
        }

        private void DrawHeader()
        {
            float tw = InsideWidth();
            if (tw < this.Size.Width)
                tw = this.Size.Width;

            GUILayout.Label("", GUIStyle.none, GUILayout.Width(tw), GUILayout.Height(InsideHeight()));
            GUILayout.BeginArea(new Rect(0, 0, tw, _headerHeight), _styleHeader);

            float x = 0;
            int i = 0;
            foreach (TableColumn col in Columns)
            {
                if (i == Columns.Count - 1)
                {
                    // Last column => auto expand
                    int expected = (int)(this.Size.Width - x);
                    if (expected < 10)
                        expected = 10;
                    if (col.Width != expected)
                        col.Width = expected;
                }

                float w = col.Width - col.Style.padding.left - col.Style.padding.right;
                float h = this._headerHeight - col.Style.padding.top - col.Style.padding.bottom;
                Rect r = new Rect(x, 0, w, h);
                GUILayout.BeginArea(r);

                GUILayout.Label(col.Caption, col.Style, GUILayout.Width(w), GUILayout.Height(h));
                GUILayout.EndArea();

                x += w;
                i++;
            }

            GUILayout.EndArea();
        }

        private void DrawRows()
        {
            float tw = InsideWidth();
            if (tw < this.Size.Width)
                tw = this.Size.Width;

            float y = this._headerHeight;
            bool even = true;
            int rowId = 0;
            foreach (TableRow row in this.Rows)
            {
                even = !even;

                GUIStyle s = new GUIStyle(_styleRow);
                s.fixedHeight = row.Height;
                if (rowId == _selectedRow)
                {
                    s.normal.background = selectedTexture;
                }
                else
                {
                    if (row.IsTitle)
                        s.normal.background = titleTexture;
                    else
                    {
                        if (even)
                            s.normal.background = lightTexture;
                        else
                            s.normal.background = darkTexture;
                    }
                }

                GUILayout.BeginArea(new Rect(0, y, tw, row.Height), s);
                float x = 0;
                for (int i = 0; i < Columns.Count; i++)
                {
                    TableColumn col = this.Columns[i];

                    string cell = "";
                    if (row.Cells.Count >= i + 1)
                        cell = row.Cells[i];

                    float w = col.Width - col.Style.padding.left - col.Style.padding.right;
                    float h = row.Height;
                    Rect r = new Rect(x, 0, w, h);
                    GUILayout.BeginArea(r);
                    GUIStyle style = row.Style;
                    if (rowId == _selectedRow)
                        style = _styleRowCellsSelected;
                    if (row.IsTitle)
                        style = _styleTitle;

                    if (!row.IsTitle)
                    {
                        if (GUILayout.Button(cell, style, GUILayout.Width(w), GUILayout.Height(h)))
                        {
                            // Select Row, click...
                            SelectRow(rowId);

                            if (_lastClickedRow == rowId && _lastClickTime > EditorApplication.timeSinceStartup - DOUBLE_CLICK_TIME)
                            {
                                // Double click!
                                if (OnDoubleClick != null)
                                    OnDoubleClick(this, rowId);
                            }

                            _lastClickedRow = rowId;
                            _lastClickTime = EditorApplication.timeSinceStartup;
                        }
                    }
                    else
                    {
                        GUILayout.Label(cell, style, GUILayout.Width(this.Size.Width), GUILayout.Height(h));
                    }
                    GUILayout.EndArea();

                    x += w;
                }

                y += row.Height;
                rowId++;
                GUILayout.EndArea();
            }
        }


        private void DrawInherited()
        {
            this.scrollPos = GUILayout.BeginScrollView(scrollPos);
            DrawHeader();
            DrawRows();
            GUILayout.EndScrollView();
        }

        public void AddRow()
        {
            this.Rows.Add(new TableRow(_styleRowCells));
            GetLastRow().ExpectedColCount = this.Columns.Count;
        }

        public TableRow GetLastRow()
        {
            if (this.Rows.Count == 0)
                return null;
            return this.Rows[this.Rows.Count - 1];
        }

        public void AddRow(string[] values)
        {
            AddRow();
            foreach (string v in values)
            {
                GetLastRow().AddValue(v);
            }
        }

        public void AddRow(List<string> values)
        {
            AddRow(values.ToArray());
        }

        public void AddTitle(string title)
        {
            AddRow();
            GetLastRow().AddValue(title);
            GetLastRow().IsTitle = true;
            GetLastRow().Height = 28;
        }

    }

}